@@ -5,6 +5,8 @@ Blind reimplementation of WebSockets as a standalone wrapper for Twisted
 protocols.
 """
 
+__version__ = "0.6.2"
+
 from base64 import b64encode, b64decode
 from hashlib import md5, sha1
 from string import digits
@@ -15,13 +17,23 @@ from twisted.protocols.policies import ProtocolWrapper, WrappingFactory
 from twisted.python import log
 from twisted.web.http import datetimeToString
 
+class WSException(Exception):
+    """
+    Something stupid happened here.
+
+    If this class escapes txWS, then something stupid happened in multiple
+    places.
+    """
+
 # Flavors of WS supported here.
 # HYBI00 - Hixie-76, HyBi-00. Challenge/response after headers, very minimal
 #          framing. Tricky to start up, but very smooth sailing afterwards.
 # HYBI07 - HyBi-07. Modern "standard" handshake. Bizarre masked frames, lots
 #          of binary data packing.
+# HYBI10 - HyBi-10. Just like HyBi-07. No, seriously. *Exactly* the same,
+#          except for the protocol number.
 
-HYBI00, HYBI07 = range(2)
+HYBI00, HYBI07, HYBI10 = range(3)
 
 # States of the state machine. Because there are no reliable byte counts for
 # any of this, we don't use StatefulProtocol; instead, we use custom state
@@ -139,21 +151,23 @@ def parse_hybi00_frames(buf):
     """
 
     start = buf.find("\x00")
+    tail = 0
     frames = []
 
     while start != -1:
-        end = buf.find("\xff")
+        end = buf.find("\xff", start + 1)
         if end == -1:
             # Incomplete frame, try again later.
-            return frames, buf
+            break
         else:
-            frame, buf = buf[start + 1:end], buf[end + 1:]
             # Found a frame, put it in the list.
+            frame = buf[start + 1:end]
             frames.append((NORMAL, frame))
-        start = buf.find("\x00")
+            tail = end + 1
+        start = buf.find("\x00", end + 1)
 
-    # We appear to have hit an exact frame boundary. This is actually pretty
-    # likely with a lot of implementations.
+    # Adjust the buffer and return.
+    buf = buf[tail:]
     return frames, buf
 
 def mask(buf, key):
@@ -170,7 +184,7 @@ def mask(buf, key):
         buf[i] = chr(ord(char) ^ key[i % 4])
     return "".join(buf)
 
-def make_hybi07_frame(buf):
+def make_hybi07_frame(buf, opcode=0x1):
     """
     Make a HyBi-07 frame.
 
@@ -186,26 +200,29 @@ def make_hybi07_frame(buf):
         length = chr(len(buf))
 
     # Always make a normal packet.
-    frame = "\x81%s%s" % (length, buf)
+    header = chr(0x80 | opcode)
+    frame = "%s%s%s" % (header, length, buf)
     return frame
 
 def parse_hybi07_frames(buf):
     """
     Parse HyBi-07 frames in a highly compliant manner.
-
-    This function requires *complete* frames and does highly bogus things if
-    fed incomplete frames.
     """
 
+    start = 0
     frames = []
 
-    while buf:
+    while True:
+        # If there's not at least two bytes in the buffer, bail.
+        if len(buf) - start < 2:
+            break
+
         # Grab the header. This single byte holds some flags nobody cares
         # about, and an opcode which nobody cares about.
-        header, buf = ord(buf[0]), buf[1:]
+        header = ord(buf[start])
         if header & 0x70:
             # At least one of the reserved flags is set. Pork chop sandwiches!
-            raise Exception("Reserved flag in HyBi-07 frame (%d)" % header)
+            raise WSException("Reserved flag in HyBi-07 frame (%d)" % header)
             frames.append(("", CLOSE))
             return frames, buf
 
@@ -215,42 +232,67 @@ def parse_hybi07_frames(buf):
         try:
             opcode = opcode_types[opcode]
         except KeyError:
-            raise Exception("Unknown opcode %d in HyBi-07 frame!" % opcode)
+            raise WSException("Unknown opcode %d in HyBi-07 frame" % opcode)
 
         # Get the payload length and determine whether we need to look for an
         # extra length.
-        length, buf = ord(buf[0]), buf[1:]
+        length = ord(buf[start + 1])
         masked = length & 0x80
         length &= 0x7f
 
+        # The offset we're gonna be using to walk through the frame. We use
+        # this because the offset is variable depending on the length and
+        # mask.
+        offset = 2
+
         # Extra length fields.
         if length == 0x7e:
-            length, buf = buf[:4], buf[4:]
+            if len(buf) - start < 4:
+                break
+
+            length = buf[start + 2:start + 4]
             length = unpack(">H", length)[0]
+            offset += 2
         elif length == 0x7f:
+            if len(buf) - start < 10:
+                break
+
             # Protocol bug: The top bit of this long long *must* be cleared;
             # that is, it is expected to be interpreted as signed. That's
             # fucking stupid, if you don't mind me saying so, and so we're
             # interpreting it as unsigned anyway. If you wanna send exabytes
             # of data down the wire, then go ahead!
-            length, buf = buf[:8], buf[8:]
+            length = buf[start + 2:start + 10]
             length = unpack(">Q", length)[0]
+            offset += 8
 
         if masked:
-            key, buf = buf[:4], buf[4:]
+            if len(buf) - (start + offset) < 4:
+                break
 
-        data, buf = buf[:length], buf[length:]
+            key = buf[start + offset:start + offset + 4]
+            offset += 4
+
+        if len(buf) - (start + offset) < length:
+            break
+
+        data = buf[start + offset:start + offset + length]
 
         if masked:
             data = mask(data, key)
 
         if opcode == CLOSE:
-            # Gotta unpack the opcode and return usable data here.
-            data = unpack(">H", data[:2])[0], data[2:]
+            if len(data) >= 2:
+                # Gotta unpack the opcode and return usable data here.
+                data = unpack(">H", data[:2])[0], data[2:]
+            else:
+                # No reason given; use generic data.
+                data = 1000, "No reason given"
 
         frames.append((opcode, data))
+        start += offset + length
 
-    return frames, buf
+    return frames, buf[start:]
 
 class WebSocketProtocol(ProtocolWrapper):
     """
@@ -329,12 +371,17 @@ class WebSocketProtocol(ProtocolWrapper):
 
         if self.flavor == HYBI00:
             parser = parse_hybi00_frames
-        elif self.flavor == HYBI07:
+        elif self.flavor in (HYBI07, HYBI10):
             parser = parse_hybi07_frames
         else:
-            raise Exception("Unknown flavor %r!" % self.flavor)
+            raise WSException("Unknown flavor %r" % self.flavor)
 
-        frames, self.buf = parser(self.buf)
+        try:
+            frames, self.buf = parser(self.buf)
+        except WSException, wse:
+            # Couldn't parse all the frames, something went wrong, let's bail.
+            self.close(wse.args[0])
+            return
 
         for frame in frames:
             opcode, data = frame
@@ -349,9 +396,8 @@ class WebSocketProtocol(ProtocolWrapper):
                 reason, text = data
                 log.msg("Closing connection: %r (%d)" % (text, reason))
 
-                # Send the smallest possible closing message.
-                # XXX totally protocol-7 specific
-                self.transport.write("\x88\x00")
+                # Close the connection.
+                self.close()
 
     def sendFrames(self):
         """
@@ -363,10 +409,10 @@ class WebSocketProtocol(ProtocolWrapper):
 
         if self.flavor == HYBI00:
             maker = make_hybi00_frame
-        elif self.flavor == HYBI07:
+        elif self.flavor in (HYBI07, HYBI10):
             maker = make_hybi07_frame
         else:
-            raise Exception("Unknown flavor %r!" % self.flavor)
+            raise WSException("Unknown flavor %r" % self.flavor)
 
         for frame in self.pending_frames:
             # Encode the frame before sending it.
@@ -421,6 +467,11 @@ class WebSocketProtocol(ProtocolWrapper):
                 self.sendHyBi07Preamble()
                 self.flavor = HYBI07
                 self.state = FRAMES
+            elif version == "8":
+                log.msg("Starting HyBi-10 conversation")
+                self.sendHyBi07Preamble()
+                self.flavor = HYBI10
+                self.state = FRAMES
             else:
                 log.msg("Can't support protocol version %s!" % version)
                 return False
@@ -469,13 +520,20 @@ class WebSocketProtocol(ProtocolWrapper):
                     self.sendHyBi00Preamble()
                     self.transport.write(response)
                     log.msg("Completed HyBi-00/Hixie-76 handshake")
-                    # Start sending frames, and kick any pending frames.
+                    # We're all finished here; start sending frames.
                     self.state = FRAMES
-                    self.sendFrames()
 
             elif self.state == FRAMES:
                 self.parseFrames()
 
+        # Kick any pending frames. This is needed because frames might have
+        # started piling up early; we can get write()s from our protocol above
+        # when they makeConnection() immediately, before our browser client
+        # actually sends any data. In those cases, we need to manually kick
+        # pending frames.
+        if self.pending_frames:
+            self.sendFrames()
+
     def write(self, data):
         """
         Write to the transport.
@@ -486,6 +544,24 @@ class WebSocketProtocol(ProtocolWrapper):
         self.pending_frames.append(data)
         self.sendFrames()
 
+    def close(self, reason=""):
+        """
+        Close the connection.
+
+        This includes telling the other side we're closing the connection.
+
+        If the other side didn't signal that the connection is being closed,
+        then we might not see their last message, but since their last message
+        should, according to the spec, be a simple acknowledgement, it
+        shouldn't be a problem.
+        """
+
+        if self.flavor in (HYBI07, HYBI10):
+            frame = make_hybi07_frame(reason, opcode=0x8)
+            self.transport.write(frame)
+
+        self.loseConnection()
+
 class WebSocketFactory(WrappingFactory):
     """
     Factory which wraps another factory to provide WebSockets transports for
